from pydantic import BaseModel


class Cards(BaseModel):
    quantities: dict[str, int]


class Player(BaseModel):
    name: str
    hand: Cards | None


class Game(BaseModel):
    finished: bool
    players: list[Player]
    stock: Cards



import logging
from typing import TYPE_CHECKING

from dopynion.cards import Card, CardContainer, CardName
from dopynion.data_model import Player as PlayerData
from dopynion.exceptions import (
    ActionDuringBuyError,
    InvalidActionError,
    InvalidBuyError,
    NotEnoughMoneyError,
)

if TYPE_CHECKING:
    import dopynion.game


class State:
    ACTION = 1
    BUY = 2
    ADJUST = 3


class Player:
    def __init__(self, name: str) -> None:
        self.game: dopynion.game.Game = None
        self.name = name
        self.deck = CardContainer()
        self.deck.append_several(7, CardName.COPPER)
        self.deck.append_several(3, CardName.ESTATE)
        self.deck.shuffle()
        self.hand = CardContainer()
        self.discard = CardContainer()
        self.played_cards = CardContainer()
        self.actions_left: int = 0
        self.purchases_left: int = 0
        self.money: int = 0
        self.playing = False
        self.state_machine = State.ACTION
        self._adjust()

    def __repr__(self) -> str:
        ret = f"Player: {self.name}\n"
        ret += f" - actions left: {self.actions_left}, "
        ret += f"purchases left: {self.purchases_left}, "
        ret += f"money left: {self.money + self.hand.money}\n"
        ret += f" - hand: {self.hand}\n"
        ret += f" - played cards: {self.played_cards}\n"
        ret += f" - deck: {self.deck}\n"
        ret += f" - discard: {self.discard}\n"
        return ret

    def _check_for_action_to_buy_transition(self) -> None:
        if not self.hand.contains_action() or not self.actions_left:
            self.actions_left = 0
            self.state_machine = State.BUY
            logging.debug("go from action to buy")

    def _check_for_buy_to_adjust_transition(self) -> None:
        if not self.hand.contains_money() or not self.purchases_left:
            self.purchases_left = 0
            self.state_machine = State.ADJUST
            logging.debug("go from buy to adjust")

    def start_turn(self) -> None:
        logging.info("start turn (%s, %d)", self.name, id(self))
        self.playing = True
        self.state_machine = State.ACTION
        self.actions_left = 1
        self.purchases_left = 1
        self.money = 0
        self._check_for_action_to_buy_transition()

    def end_turn(self) -> None:
        self._adjust()
        self.playing = False
        logging.info("end turn")

    def take_one_card_from_deck(self) -> CardName:
        if not self.deck:
            self.discard.empty_to(self.deck)
            self.deck.shuffle()
        return self.deck.pop(0)

    def _adjust(self) -> None:
        self.played_cards.empty_to(self.discard)
        self.hand.empty_to(self.discard)
        for _ in range(5):
            self.hand.append(self.take_one_card_from_deck())
        self.state_machine = State.ADJUST

    def _prepare_money(self, money: int) -> None:
        money_cards = self.hand.money_cards
        money_cards.sort(key=lambda card_name: Card.types[card_name].money)
        while self.money < money:
            if not money_cards:
                break
            money_card = money_cards.pop(0)
            self.money += Card.types[money_card].money
            self.hand.remove(money_card)
            self.played_cards.append(money_card)

    def buy(self, card_name: CardName) -> None:
        logging.info("> BUY %s", card_name)
        quantity = getattr(self.game.stock, card_name + "_qty")
        if not quantity:
            raise InvalidBuyError(card_name)
        card = Card.types[card_name]
        if self.money + self.hand.money < card.cost:
            raise NotEnoughMoneyError
        self._prepare_money(card.cost)
        self.money -= card.cost
        self.game.stock.remove(card_name)
        self.discard.append(card_name)
        self.purchases_left -= 1
        self._check_for_buy_to_adjust_transition()

    def action(self, card_name: CardName) -> None:
        logging.info("> ACTION %s", card_name)
        if self.state_machine != State.ACTION:
            logging.debug(self.state_machine)
            logging.debug("actions_left %d", self.actions_left)
            logging.debug(self.state)
            raise ActionDuringBuyError(card_name)
        if card_name not in self.hand:
            logging.debug(self.state)
            raise InvalidActionError(card_name)
        try:
            Card.types[card_name].action(self)
        except NotImplementedError as err:
            raise InvalidActionError(card_name) from err
        self.actions_left -= 1
        self.hand.remove(card_name)
        self.played_cards.append(card_name)
        self._check_for_action_to_buy_transition()

    @property
    def state(self) -> PlayerData:
        state = PlayerData(name=self.name, hand=None)
        if self.playing:
            state.hand = self.hand.state
        return state



class InvalidCommandError(Exception):
    pass


class AddPlayerDuringGameError(InvalidCommandError):
    pass


class InvalidActionError(InvalidCommandError):
    pass


class ActionDuringBuyError(InvalidCommandError):
    pass


class MissingCardError(InvalidCommandError):
    pass


class InvalidBuyError(InvalidCommandError):
    pass


class NotEnoughMoneyError(InvalidCommandError):
    pass



import pytest

from dopynion.cards import CardName
from dopynion.exceptions import (
    ActionDuringBuyError,
    InvalidActionError,
)
from dopynion.player import Player, State


def test_initial_deck() -> None:
    player = Player("Toto")
    assert len(player.deck) + len(player.hand) == 10
    assert player.deck.copper_qty + player.hand.copper_qty == 7
    assert player.deck.estate_qty + player.hand.estate_qty == 3


def test_make_hand_enough_cards_in_deck() -> None:
    player = Player("toto")
    assert len(player.hand) == 5
    assert len(player.deck) == 5


def test_make_hand_not_enough_cards_in_deck() -> None:
    pass  # TODO add the test


def test_unknown_action() -> None:
    player = Player("toto")
    player.start_turn()
    player.state_machine = State.ACTION
    with pytest.raises(InvalidActionError):
        player.action("FooBar")


def test_invalid_action() -> None:
    player = Player("toto")
    player.start_turn()
    with pytest.raises(ActionDuringBuyError):
        player.action(CardName.SMITHY)


def test_invalid_action_when_no_corresponding_card() -> None:
    player = Player("toto")
    player.start_turn()
    player.state_machine = State.ACTION
    with pytest.raises(InvalidActionError):
        player.action(CardName.VILLAGE)



from __future__ import annotations

import inspect
import random
import sys
from collections import defaultdict
from enum import StrEnum
from typing import TYPE_CHECKING, ClassVar

from dopynion.data_model import Cards

if TYPE_CHECKING:
    from collections.abc import Callable

    from dopynion.player import Player


class CardName(StrEnum):  # Create with a metaclass
    COPPER = "copper"
    COUNCILROOM = "councilroom"
    CURSE = "curse"
    DUCHY = "duchy"
    ESTATE = "estate"
    FESTIVAL = "festival"
    GOLD = "gold"
    LABORATORY = "laboratory"
    MARKET = "market"
    PROVINCE = "province"
    SILVER = "silver"
    SMITHY = "smithy"
    VILLAGE = "village"
    WOODCUTTER = "woodcutter"

    def __repr__(self) -> str:
        return self.value.title()


class ClassNameRepr(type):
    def __repr__(cls) -> str:
        return cls.__name__


class Card(metaclass=ClassNameRepr):
    types: ClassVar[dict[CardName, type[Card]]] = {}
    name = "Unknown"
    cost = 10_000
    money = 0
    is_action = False
    is_kingdom = True
    is_money = False

    def __init_subclass__(cls) -> None:
        Card.types[CardName[cls.__name__.upper()]] = cls

    def __eq__(self, other: object) -> bool:
        return isinstance(self, other) or self.__class__ is other

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return self.name

    @classmethod
    def action(cls, player: Player) -> None:
        raise NotImplementedError


class Copper(Card):
    name = "Cuivre"
    cost = 0
    money = 1
    is_kingdom = False
    is_money = True


class CouncilRoom(Card):
    name = "Chambre du conseil"
    cost = 5
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        for _ in range(4):
            player.hand.append(player.take_one_card_from_deck())
        player.purchases_left += 1


class Curse(Card):
    name = "Malédiction"
    cost = 0
    is_kingdom = False


class Duchy(Card):
    name = "Duché"
    cost = 5
    is_kingdom = False


class Estate(Card):
    name = "Domaine"
    cost = 2
    is_kingdom = False


class Festival(Card):
    name = "Festival"
    cost = 5
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        player.actions_left += 2
        player.purchases_left += 1
        player.money += 2


class Gold(Card):
    name = "Or"
    cost = 6
    money = 3
    is_kingdom = False
    is_money = True


class Laboratory(Card):
    name = "Laboratoire"
    cost = 5
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        for _ in range(2):
            player.hand.append(player.take_one_card_from_deck())
        player.actions_left += 1


class Market(Card):
    name = "Marché"
    cost = 5
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        player.hand.append(player.take_one_card_from_deck())
        player.actions_left += 1
        player.purchases_left += 1
        player.money += 1


class Province(Card):
    name = "Province"
    cost = 8
    is_kingdom = False


class Silver(Card):
    name = "Argent"
    cost = 4
    money = 2
    is_kingdom = False
    is_money = True


class Smithy(Card):
    name = "Forgeron"
    cost = 4
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        for _ in range(3):
            player.hand.append(player.take_one_card_from_deck())


class Village(Card):
    name = "Village"
    cost = 3
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        player.hand.append(player.take_one_card_from_deck())
        player.actions_left += 2


class Woodcutter(Card):
    name = "Bucheron"
    cost = 3
    is_action = True

    @classmethod
    def action(cls, player: Player) -> None:
        player.purchases_left += 1
        player.money += 2


actions_card_name: set[CardName] = {
    name.lower()
    for name, class_ in inspect.getmembers(sys.modules[__name__], inspect.isclass)
    if issubclass(class_, Card) and class_.name != "Unknown" and class_.is_action
}

money_card_name: set[CardName] = {CardName.COPPER, CardName.SILVER, CardName.GOLD}


class CardContainer:
    """Storage for card."""

    def __init__(self) -> None:
        self._quantities: dict[CardName, int] = defaultdict(int)
        self._cards: list[CardName] = []

    def __getattr__(self, attribute: str) -> int:
        if attribute.endswith("_qty"):
            card_name = attribute[:-4]
            return self._quantities[CardName[card_name.upper()]]
        raise AttributeError

    def __repr__(self) -> str:
        return str(self._cards)

    def append(self, card_name: CardName) -> None:
        self._cards.append(card_name)
        self._quantities[card_name] += 1

    def append_several(self, qty: int, card_name: CardName) -> None:
        for _ in range(qty):
            self.append(card_name)

    def remove(self, card_name: CardName) -> None:
        self._quantities[card_name] -= 1
        self._cards.remove(card_name)

    def shuffle(self) -> None:
        random.shuffle(self._cards)

    def contains_action(self) -> bool:
        return any(
            card_name in actions_card_name
            for card_name, qty in self._quantities.items()
            if qty > 0
        )

    @property
    def action_cards(self) -> CardContainer:
        ret = CardContainer()
        for card_name, qty in self._quantities.items():
            if card_name in actions_card_name:
                ret.append_several(qty, card_name)
        return ret

    def contains_money(self) -> bool:
        return any(
            card_name in money_card_name
            for card_name, qty in self._quantities.items()
            if qty > 0
        )

    @property
    def money_cards(self) -> CardContainer:
        ret = CardContainer()
        for card_name, qty in self._quantities.items():
            if card_name in money_card_name:
                ret.append_several(qty, card_name)
        return ret

    @property
    def money(self) -> int:
        return (
            1 * self._quantities.get(CardName.COPPER, 0)
            + 2 * self._quantities.get(CardName.SILVER, 0)
            + 3 * self._quantities.get(CardName.GOLD, 0)
        )

    def pop(self, index: int = -1) -> CardName:
        card_name = self._cards.pop(index)
        self._quantities[card_name] -= 1
        return card_name

    def __contains__(self, card_name: CardName) -> bool:
        return self._quantities.get(card_name, 0) > 0

    def __len__(self) -> int:
        return len(self._cards)

    def __getitem__(self, index: int) -> CardName:
        return self._cards[index]

    def quantity(self, card_name: CardName) -> int:
        return self._quantities.get(card_name, 0)

    def sort(self, key: Callable, *, reverse: bool = False) -> None:
        self._cards.sort(key=key, reverse=reverse)

    def empty_to(self, other: CardContainer) -> None:
        while self:
            card_name = self.pop(0)
            other.append(card_name)
        self._quantities.clear()
        self._cards.clear()

    @property
    def three_empty_piles(self) -> bool:
        nb_piles = 3
        return sum(1 for qty in self._quantities.values() if qty == 0) >= nb_piles

    @property
    def state(self) -> Cards:
        return Cards(quantities=self._quantities.copy())



# example of execution

import logging
import random

from dopynion.cards import Card, CardName
from dopynion.game import Game
from dopynion.player import Player

logging.basicConfig(
    filename="game.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format=(
        "%(asctime)s [%(levelname)-8s] %(filename)20s(%(lineno)3s):"
        "%(funcName)-20s :: %(message)s"
    ),
    datefmt="%m/%d/%Y %H:%M:%S",
)
logging.info("Launching server")

game = Game()
player1 = Player("Foo")
player2 = Player("Bar")
player3 = Player("Baz")
game.add_player(player1)
game.add_player(player2)
game.add_player(player3)
game.start()

turn = 0
while not game.finished:
    turn += 1
    print("#" * 80)
    print(f"# turn {turn}")
    print("#" * 80)
    print(game.state)
    player1.start_turn()
    print(player1.state)

    while player1.actions_left and player1.hand.contains_action():
        actions = player1.hand.action_cards
        player1.action(random.choice(actions))  # noqa: S311
        print(player1)

    while player1.purchases_left and player1.hand.contains_money():
        money = player1.hand.money
        buyables: list[CardName] = [
            card_name
            for card_name, class_ in Card.types.items()
            if class_.cost <= money and card_name in game.stock
        ]
        player1.buy(random.choice(buyables))  # noqa: S311
        print(player1)

    player1.end_turn()
